@@ -3,7 +3,7 @@
  * @author Michal Vasko <mvasko@cesnet.cz>
  * @brief LYB data parser for libyang
  *
- * Copyright (c) 2020 CESNET, z.s.p.o.
+ * Copyright (c) 2020 - 2022 CESNET, z.s.p.o.
  *
  * This source code is licensed under BSD 3-Clause License (the "License").
  * You may not use this file except in compliance with the License.
@@ -31,6 +31,7 @@
 #include "parser_data.h"
 #include "parser_internal.h"
 #include "plugins_exts.h"
+#include "plugins_exts/metadata.h"
 #include "set.h"
 #include "tree.h"
 #include "tree_data.h"
@@ -40,22 +41,23 @@
 #include "validation.h"
 #include "xml.h"
 
-static LY_ERR _lyd_parse_lyb(const struct ly_ctx *ctx, const struct lysc_ext_instance *ext, struct lyd_node *parent,
-        struct lyd_node **first_p, struct ly_in *in, uint32_t parse_opts, uint32_t val_opts, uint32_t int_opts,
-        struct ly_set *parsed, struct lyd_ctx **lydctx_p);
-
-static LY_ERR lyb_parse_siblings(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, struct lyd_node **first_p, struct ly_set *parsed);
+static LY_ERR lyb_parse_siblings(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, struct lyd_node **first_p,
+        struct ly_set *parsed);
 
 void
 lylyb_ctx_free(struct lylyb_ctx *ctx)
 {
     LY_ARRAY_COUNT_TYPE u;
 
+    if (!ctx) {
+        return;
+    }
+
     LY_ARRAY_FREE(ctx->siblings);
     LY_ARRAY_FREE(ctx->models);
 
     LY_ARRAY_FOR(ctx->sib_hts, u) {
-        lyht_free(ctx->sib_hts[u].ht);
+        lyht_free(ctx->sib_hts[u].ht, NULL);
     }
     LY_ARRAY_FREE(ctx->sib_hts);
 
@@ -67,6 +69,10 @@ lyd_lyb_ctx_free(struct lyd_ctx *lydctx)
 {
     struct lyd_lyb_ctx *ctx = (struct lyd_lyb_ctx *)lydctx;
 
+    if (!lydctx) {
+        return;
+    }
+
     lyd_ctx_free(lydctx);
     lylyb_ctx_free(ctx->lybctx);
     free(ctx);
@@ -337,12 +343,14 @@ lyb_read_start_siblings(struct lylyb_ctx *lybctx)
  * @param[in] lybctx LYB context.
  * @param[out] mod_name Module name, if any.
  * @param[out] mod_rev Module revision, "" if none.
+ * @param[in,out] feat_set Set to add the names of enabled features to. If not set, enabled features are not parsed.
  * @return LY_ERR value.
  */
 static LY_ERR
-lyb_read_model(struct lylyb_ctx *lybctx, char **mod_name, char mod_rev[])
+lyb_read_model(struct lylyb_ctx *lybctx, char **mod_name, char mod_rev[], struct ly_set *feat_set)
 {
-    uint16_t rev, length;
+    uint16_t i, rev, length;
+    char *str;
 
     *mod_name = NULL;
     mod_rev[0] = '\0';
@@ -366,6 +374,23 @@ lyb_read_model(struct lylyb_ctx *lybctx, char **mod_name, char mod_rev[])
                 (rev & LYB_REV_MONTH_MASK) >> LYB_REV_MONTH_SHIFT, rev & LYB_REV_DAY_MASK);
     }
 
+    if (!feat_set) {
+        /* enabled features not printed */
+        return LY_SUCCESS;
+    }
+
+    /* enabled feature count */
+    lyb_read_number(&length, sizeof length, sizeof length, lybctx);
+    if (!length) {
+        return LY_SUCCESS;
+    }
+
+    /* enabled features */
+    for (i = 0; i < length; ++i) {
+        LY_CHECK_RET(lyb_read_string(&str, sizeof length, lybctx));
+        ly_set_add(feat_set, str, 1, NULL);
+    }
+
     return LY_SUCCESS;
 }
 
@@ -374,18 +399,23 @@ lyb_read_model(struct lylyb_ctx *lybctx, char **mod_name, char mod_rev[])
  *
  * @param[in] lybctx LYB context.
  * @param[in] parse_options Flag with options for parsing.
+ * @param[in] with_features Whether the enabled features were also printed and should be read.
  * @param[out] mod Parsed module.
  * @return LY_ERR value.
  */
 static LY_ERR
-lyb_parse_model(struct lylyb_ctx *lybctx, uint32_t parse_options, const struct lys_module **mod)
+lyb_parse_model(struct lylyb_ctx *lybctx, uint32_t parse_options, ly_bool with_features, const struct lys_module **mod)
 {
     LY_ERR ret = LY_SUCCESS;
     const struct lys_module *m = NULL;
     char *mod_name = NULL, mod_rev[LY_REV_SIZE];
+    struct ly_set feat_set = {0};
+    struct lysp_feature *f = NULL;
+    uint32_t i, idx = 0;
+    ly_bool enabled;
 
     /* read module info */
-    if ((ret = lyb_read_model(lybctx, &mod_name, mod_rev))) {
+    if ((ret = lyb_read_model(lybctx, &mod_name, mod_rev, with_features ? &feat_set : NULL))) {
         goto cleanup;
     }
 
@@ -417,16 +447,41 @@ lyb_parse_model(struct lylyb_ctx *lybctx, uint32_t parse_options, const struct l
             goto cleanup;
         }
 
+        goto cleanup;
     }
 
-    if (m) {
-        /* fill cached hashes, if not already */
-        lyb_cache_module_hash(m);
+    if (with_features) {
+        /* check features */
+        while ((f = lysp_feature_next(f, m->parsed, &idx))) {
+            enabled = 0;
+            for (i = 0; i < feat_set.count; ++i) {
+                if (!strcmp(feat_set.objs[i], f->name)) {
+                    enabled = 1;
+                    break;
+                }
+            }
+
+            if (enabled && !(f->flags & LYS_FENABLED)) {
+                LOGERR(lybctx->ctx, LY_EINVAL, "Invalid context for LYB data parsing, module \"%s\" has \"%s\" feature disabled.",
+                        mod_name, f->name);
+                ret = LY_EINVAL;
+                goto cleanup;
+            } else if (!enabled && (f->flags & LYS_FENABLED)) {
+                LOGERR(lybctx->ctx, LY_EINVAL, "Invalid context for LYB data parsing, module \"%s\" has \"%s\" feature enabled.",
+                        mod_name, f->name);
+                ret = LY_EINVAL;
+                goto cleanup;
+            }
+        }
     }
 
+    /* fill cached hashes, if not already */
+    lyb_cache_module_hash(m);
+
 cleanup:
     *mod = m;
     free(mod_name);
+    ly_set_erase(&feat_set, free);
     return ret;
 }
 
@@ -453,7 +508,7 @@ lyb_parse_metadata(struct lyd_lyb_ctx *lybctx, const struct lysc_node *sparent,
     /* read attributes */
     for (i = 0; i < count; ++i) {
         /* find model */
-        ret = lyb_parse_model(lybctx->lybctx, lybctx->parse_opts, &mod);
+        ret = lyb_parse_model(lybctx->lybctx, lybctx->parse_opts, 0, &mod);
         LY_CHECK_GOTO(ret, cleanup);
 
         if (!mod) {
@@ -856,46 +911,6 @@ lyb_skip_siblings(struct lylyb_ctx *lybctx)
     } while (LYB_LAST_SIBLING(lybctx).written);
 }
 
-/**
- * @brief Parse the context of anydata/anyxml node.
- *
- * @param[in] ctx libyang context.
- * @param[in] data LYB data to parse.
- * @param[out] tree Parsed tree.
- * @return LY_ERR value.
- */
-static LY_ERR
-lyb_parse_any_content(const struct ly_ctx *ctx, const char *data, struct lyd_node **tree)
-{
-    LY_ERR ret;
-    uint32_t prev_lo;
-    struct ly_in *in;
-    struct lyd_ctx *lydctx = NULL;
-
-    *tree = NULL;
-
-    LY_CHECK_RET(ly_in_new_memory(data, &in));
-
-    /* turn logging off */
-    prev_lo = ly_log_options(0);
-
-    ret = _lyd_parse_lyb(ctx, NULL, NULL, tree, in, LYD_PARSE_ONLY | LYD_PARSE_OPAQ | LYD_PARSE_STRICT, 0,
-            LYD_INTOPT_ANY | LYD_INTOPT_WITH_SIBLINGS, NULL, &lydctx);
-
-    /* turn logging on again */
-    ly_log_options(prev_lo);
-
-    ly_in_free(in, 0);
-    if (lydctx) {
-        lydctx->free(lydctx);
-    }
-    if (ret) {
-        lyd_free_siblings(*tree);
-        *tree = NULL;
-    }
-    return ret;
-}
-
 /**
  * @brief Insert new node to @p parsed set.
  *
@@ -914,7 +929,7 @@ lyb_insert_node(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, struct lyd_
 {
     /* insert, keep first pointer correct */
     if (parent && (LYD_CTX(parent) != LYD_CTX(node))) {
-        lyd_insert_ext(parent, node);
+        lyplg_ext_insert(parent, node);
     } else {
         lyd_insert_node(parent, first_p, node, lybctx->parse_opts & LYD_PARSE_ORDERED ? 1 : 0);
     }
@@ -989,7 +1004,14 @@ lyb_finish_node(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, uint32_t fl
     (*node)->meta = *meta;
     *meta = NULL;
 
+    /* insert into parent */
     lyb_insert_node(lybctx, parent, *node, first_p, parsed);
+
+    if (!(lybctx->parse_opts & LYD_PARSE_ONLY)) {
+        /* store for ext instance node validation, if needed */
+        (void)lyd_validate_node_ext(*node, &lybctx->ext_node);
+    }
+
     *node = NULL;
 }
 
@@ -1068,13 +1090,13 @@ lyb_validate_node_inner(struct lyd_lyb_ctx *lybctx, const struct lysc_node *snod
 
     if (!(lybctx->parse_opts & LYD_PARSE_ONLY)) {
         /* new node validation, autodelete CANNOT occur, all nodes are new */
-        ret = lyd_validate_new(lyd_node_child_p(node), snode, NULL, NULL);
+        ret = lyd_validate_new(lyd_node_child_p(node), snode, NULL, 0, NULL);
         LY_CHECK_RET(ret);
 
         /* add any missing default children */
         impl_opts = (lybctx->val_opts & LYD_VALIDATE_NO_STATE) ? LYD_IMPLICIT_NO_STATE : 0;
         ret = lyd_new_implicit_r(node, lyd_node_child_p(node), NULL, NULL, &lybctx->node_when, &lybctx->node_types,
-                impl_opts, NULL);
+                &lybctx->ext_node, impl_opts, NULL);
         LY_CHECK_RET(ret);
     }
 
@@ -1185,9 +1207,11 @@ lyb_parse_node_any(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, const st
         struct lyd_node **first_p, struct ly_set *parsed)
 {
     LY_ERR ret;
-    struct lyd_node *node = NULL, *tree;
+    struct lyd_node *node = NULL, *tree = NULL;
     struct lyd_meta *meta = NULL;
     LYD_ANYDATA_VALUETYPE value_type;
+    struct ly_in *in;
+    struct lyd_ctx *lydctx = NULL;
     char *value = NULL;
     uint32_t flags;
     const struct ly_ctx *ctx = lybctx->lybctx->ctx;
@@ -1198,8 +1222,7 @@ lyb_parse_node_any(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, const st
 
     /* parse value type */
     lyb_read_number(&value_type, sizeof value_type, sizeof value_type, lybctx->lybctx);
-    if (value_type == LYD_ANYDATA_DATATREE) {
-        /* invalid situation */
+    if ((value_type == LYD_ANYDATA_DATATREE) || ((snode->nodetype == LYS_ANYDATA) && (value_type != LYD_ANYDATA_LYB))) {
         LOGINT(ctx);
         ret = LY_EINT;
         goto error;
@@ -1210,18 +1233,24 @@ lyb_parse_node_any(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, const st
     LY_CHECK_GOTO(ret, error);
 
     if (value_type == LYD_ANYDATA_LYB) {
-        /* try to parse LYB into a data tree */
-        if (!lyb_parse_any_content(ctx, value, &tree)) {
-            /* successfully parsed */
-            free(value);
-            value = (char *)tree;
-            value_type = LYD_ANYDATA_DATATREE;
+        /* parse LYB into a data tree */
+        LY_CHECK_RET(ly_in_new_memory(value, &in));
+        ret = lyd_parse_lyb(ctx, NULL, NULL, &tree, in, LYD_PARSE_ONLY | LYD_PARSE_OPAQ | LYD_PARSE_STRICT, 0,
+                LYD_INTOPT_ANY | LYD_INTOPT_WITH_SIBLINGS, NULL, NULL, &lydctx);
+        ly_in_free(in, 0);
+        if (lydctx) {
+            lydctx->free(lydctx);
         }
+        LY_CHECK_ERR_GOTO(ret, lyd_free_siblings(tree), error);
+
+        /* use the parsed tree as the value */
+        free(value);
+        value = (char *)tree;
+        value_type = LYD_ANYDATA_DATATREE;
     }
 
     /* create the node */
     switch (value_type) {
-    case LYD_ANYDATA_LYB:
     case LYD_ANYDATA_DATATREE:
     case LYD_ANYDATA_STRING:
     case LYD_ANYDATA_XML:
@@ -1455,7 +1484,7 @@ lyb_parse_node(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, struct lyd_n
     switch (lyb_type) {
     case LYB_NODE_TOP:
         /* top-level, read module name */
-        LY_CHECK_GOTO(ret = lyb_parse_model(lybctx->lybctx, lybctx->parse_opts, &mod), cleanup);
+        LY_CHECK_GOTO(ret = lyb_parse_model(lybctx->lybctx, lybctx->parse_opts, 0, &mod), cleanup);
 
         /* read hash, find the schema node starting from mod */
         LY_CHECK_GOTO(ret = lyb_parse_schema_hash(lybctx, NULL, mod, &snode), cleanup);
@@ -1467,7 +1496,7 @@ lyb_parse_node(struct lyd_lyb_ctx *lybctx, struct lyd_node *parent, struct lyd_n
         break;
     case LYB_NODE_EXT:
         /* ext, read module name */
-        LY_CHECK_GOTO(ret = lyb_read_model(lybctx->lybctx, &mod_name, mod_rev), cleanup);
+        LY_CHECK_GOTO(ret = lyb_read_model(lybctx->lybctx, &mod_name, mod_rev, NULL), cleanup);
 
         /* read schema node name, find the nexted ext schema node */
         LY_CHECK_GOTO(ret = lyb_parse_schema_nested_ext(lybctx, parent, mod_name, &snode), cleanup);
@@ -1550,7 +1579,7 @@ lyb_parse_data_models(struct lylyb_ctx *lybctx, uint32_t parse_options)
 
         /* read modules */
         for (u = 0; u < count; ++u) {
-            ret = lyb_parse_model(lybctx, parse_options, &lybctx->models[u]);
+            ret = lyb_parse_model(lybctx, parse_options, 1, &lybctx->models[u]);
             LY_CHECK_RET(ret);
             LY_ARRAY_INCREMENT(lybctx->models);
         }
@@ -1614,10 +1643,10 @@ lyb_parse_header(struct lylyb_ctx *lybctx)
     return LY_SUCCESS;
 }
 
-static LY_ERR
-_lyd_parse_lyb(const struct ly_ctx *ctx, const struct lysc_ext_instance *ext, struct lyd_node *parent,
+LY_ERR
+lyd_parse_lyb(const struct ly_ctx *ctx, const struct lysc_ext_instance *ext, struct lyd_node *parent,
         struct lyd_node **first_p, struct ly_in *in, uint32_t parse_opts, uint32_t val_opts, uint32_t int_opts,
-        struct ly_set *parsed, struct lyd_ctx **lydctx_p)
+        struct ly_set *parsed, ly_bool *subtree_sibling, struct lyd_ctx **lydctx_p)
 {
     LY_ERR rc = LY_SUCCESS;
     struct lyd_lyb_ctx *lybctx;
@@ -1625,6 +1654,12 @@ _lyd_parse_lyb(const struct ly_ctx *ctx, const struct lysc_ext_instance *ext, st
     assert(!(parse_opts & ~LYD_PARSE_OPTS_MASK));
     assert(!(val_opts & ~LYD_VALIDATE_OPTS_MASK));
 
+    LY_CHECK_ARG_RET(ctx, !(parse_opts & LYD_PARSE_SUBTREE), LY_EINVAL);
+
+    if (subtree_sibling) {
+        *subtree_sibling = 0;
+    }
+
     lybctx = calloc(1, sizeof *lybctx);
     LY_CHECK_ERR_RET(!lybctx, LOGMEM(ctx), LY_EMEM);
     lybctx->lybctx = calloc(1, sizeof *lybctx->lybctx);
@@ -1684,49 +1719,12 @@ cleanup:
     return rc;
 }
 
-LY_ERR
-lyd_parse_lyb(const struct ly_ctx *ctx, const struct lysc_ext_instance *ext, struct lyd_node *parent,
-        struct lyd_node **first_p, struct ly_in *in, uint32_t parse_opts, uint32_t val_opts, enum lyd_type data_type,
-        struct ly_set *parsed, ly_bool *subtree_sibling, struct lyd_ctx **lydctx_p)
-{
-    uint32_t int_opts;
-
-    assert(!(parse_opts & ~LYD_PARSE_OPTS_MASK));
-    assert(!(val_opts & ~LYD_VALIDATE_OPTS_MASK));
-
-    LY_CHECK_ARG_RET(ctx, !(parse_opts & LYD_PARSE_SUBTREE), LY_EINVAL);
-
-    switch (data_type) {
-    case LYD_TYPE_DATA_YANG:
-        int_opts = LYD_INTOPT_WITH_SIBLINGS;
-        break;
-    case LYD_TYPE_RPC_YANG:
-        int_opts = LYD_INTOPT_RPC | LYD_INTOPT_ACTION | LYD_INTOPT_NO_SIBLINGS;
-        break;
-    case LYD_TYPE_NOTIF_YANG:
-        int_opts = LYD_INTOPT_NOTIF | LYD_INTOPT_NO_SIBLINGS;
-        break;
-    case LYD_TYPE_REPLY_YANG:
-        int_opts = LYD_INTOPT_REPLY | LYD_INTOPT_NO_SIBLINGS;
-        break;
-    default:
-        LOGINT(ctx);
-        return LY_EINT;
-    }
-
-    if (subtree_sibling) {
-        *subtree_sibling = 0;
-    }
-    return _lyd_parse_lyb(ctx, ext, parent, first_p, in, parse_opts, val_opts, int_opts, parsed, lydctx_p);
-}
-
 LIBYANG_API_DEF int
 lyd_lyb_data_length(const char *data)
 {
     LY_ERR ret = LY_SUCCESS;
     struct lylyb_ctx *lybctx;
-    int count, i;
-    size_t len;
+    uint32_t count, feat_count, len = 0, i, j;
     uint8_t buf[LYB_SIZE_MAX];
     uint8_t zero[LYB_SIZE_BYTES] = {0};
 
@@ -1753,7 +1751,6 @@ lyd_lyb_data_length(const char *data)
     /* read all models */
     for (i = 0; i < count; ++i) {
         /* module name length */
-        len = 0;
         lyb_read_number(&len, sizeof len, 2, lybctx);
 
         /* model name */
@@ -1761,6 +1758,18 @@ lyd_lyb_data_length(const char *data)
 
         /* revision */
         lyb_read(buf, 2, lybctx);
+
+        /* enabled feature count */
+        lyb_read_number(&feat_count, sizeof feat_count, 2, lybctx);
+
+        /* enabled features */
+        for (j = 0; j < feat_count; ++j) {
+            /* feature name length */
+            lyb_read_number(&len, sizeof len, 2, lybctx);
+
+            /* feature name */
+            lyb_read(buf, len, lybctx);
+        }
     }
 
     if (memcmp(zero, lybctx->in->current, LYB_SIZE_BYTES)) {
@@ -1787,5 +1796,5 @@ cleanup:
     ly_in_free(lybctx->in, 0);
     lylyb_ctx_free(lybctx);
 
-    return ret ? -1 : count;
+    return ret ? -1 : (int)count;
 }
