@@ -1,9 +1,10 @@
 /**
  * @file printer_json.c
  * @author Radek Krejci <rkrejci@cesnet.cz>
+ * @author Michal Vasko <mvasko@cesnet.cz>
  * @brief JSON printer for libyang data structure
  *
- * Copyright (c) 2015 - 2020 CESNET, z.s.p.o.
+ * Copyright (c) 2015 - 2022 CESNET, z.s.p.o.
  *
  * This source code is licensed under BSD 3-Clause License (the "License").
  * You may not use this file except in compliance with the License.
@@ -22,6 +23,7 @@
 #include "out.h"
 #include "out_internal.h"
 #include "parser_data.h"
+#include "plugins_exts/metadata.h"
 #include "plugins_types.h"
 #include "printer_data.h"
 #include "printer_internal.h"
@@ -60,11 +62,9 @@ struct jsonpr_ctx {
 static LY_ERR json_print_node(struct jsonpr_ctx *pctx, const struct lyd_node *node);
 
 /**
- * Compare 2 nodes, despite it is regular data node or an opaq node, and
+ * @brief Compare 2 nodes, despite it is regular data node or an opaq node, and
  * decide if they corresponds to the same schema node.
  *
- * TODO: rewrite lyd_compare_single and use it instead of this
- *
  * @return 1 - matching nodes, 0 - non-matching nodes
  */
 static int
@@ -81,6 +81,7 @@ matching_node(const struct lyd_node *node1, const struct lyd_node *node2)
         /* compare node names */
         struct lyd_node_opaq *onode1 = (struct lyd_node_opaq *)node1;
         struct lyd_node_opaq *onode2 = (struct lyd_node_opaq *)node2;
+
         if ((onode1->name.name != onode2->name.name) || (onode1->name.prefix != onode2->name.prefix)) {
             return 0;
         }
@@ -199,6 +200,7 @@ json_nscmp(const struct lyd_node *node1, const struct lyd_node *node2)
     } else {
         const char *pref1 = node_prefix(node1);
         const char *pref2 = node_prefix(node2);
+
         if ((pref1 && pref2) && (pref1 == pref2)) {
             return 0;
         } else {
@@ -226,6 +228,7 @@ json_print_string(struct ly_out *out, const char *text)
     ly_write_(out, "\"", 1);
     for (i = n = 0; text[i]; i++) {
         const unsigned char ascii = text[i];
+
         if (ascii < 0x20) {
             /* control character */
             ly_print_(out, "\\u%.4X", ascii);
@@ -291,18 +294,21 @@ json_print_member2(struct jsonpr_ctx *pctx, const struct lyd_node *parent, LY_VA
 
     /* determine prefix string */
     if (name) {
-        const struct lys_module *mod;
-
         switch (format) {
         case LY_VALUE_JSON:
             module_name = name->module_name;
             break;
-        case LY_VALUE_XML:
-            mod = ly_ctx_get_module_implemented_ns(pctx->ctx, name->module_ns);
+        case LY_VALUE_XML: {
+            const struct lys_module *mod = NULL;
+
+            if (name->module_ns) {
+                mod = ly_ctx_get_module_implemented_ns(pctx->ctx, name->module_ns);
+            }
             if (mod) {
                 module_name = mod->name;
             }
             break;
+        }
         default:
             /* cannot be created */
             LOGINT_RET(pctx->ctx);
@@ -400,19 +406,21 @@ json_print_attribute(struct jsonpr_ctx *pctx, const struct lyd_node_opaq *node,
     struct lyd_attr *attr;
 
     if (wdmod) {
-        ly_print_(pctx->out, "%*s\"%s:default\":\"true\"", INDENT, wdmod->name);
+        ly_print_(pctx->out, "%*s\"%s:default\":true", INDENT, wdmod->name);
         LEVEL_PRINTED;
     }
 
     for (attr = node->attr; attr; attr = attr->next) {
-        PRINT_COMMA;
         json_print_member2(pctx, &node->node, attr->format, &attr->name, 0);
 
-        if (attr->hints & (LYD_VALHINT_BOOLEAN | LYD_VALHINT_DECNUM)) {
+        if (attr->hints & (LYD_VALHINT_STRING | LYD_VALHINT_OCTNUM | LYD_VALHINT_HEXNUM | LYD_VALHINT_NUM64)) {
+            json_print_string(pctx->out, attr->value);
+        } else if (attr->hints & (LYD_VALHINT_BOOLEAN | LYD_VALHINT_DECNUM)) {
             ly_print_(pctx->out, "%s", attr->value[0] ? attr->value : "null");
         } else if (attr->hints & LYD_VALHINT_EMPTY) {
             ly_print_(pctx->out, "[null]");
         } else {
+            /* unknown value format with no hints, use universal string */
             json_print_string(pctx->out, attr->value);
         }
         LEVEL_PRINTED;
@@ -435,7 +443,7 @@ json_print_metadata(struct jsonpr_ctx *pctx, const struct lyd_node *node, const
     struct lyd_meta *meta;
 
     if (wdmod) {
-        ly_print_(pctx->out, "%*s\"%s:default\":\"true\"", INDENT, wdmod->name);
+        ly_print_(pctx->out, "%*s\"%s:default\":true", INDENT, wdmod->name);
         LEVEL_PRINTED;
     }
 
@@ -462,15 +470,16 @@ json_print_attributes(struct jsonpr_ctx *pctx, const struct lyd_node *node, ly_b
 {
     const struct lys_module *wdmod = NULL;
 
-    if ((node->flags & LYD_DEFAULT) && (pctx->options & (LYD_PRINT_WD_ALL_TAG | LYD_PRINT_WD_IMPL_TAG))) {
+    if (node->schema && (node->schema->nodetype != LYS_CONTAINER) && (node->flags & LYD_DEFAULT) &&
+            (pctx->options & (LYD_PRINT_WD_ALL_TAG | LYD_PRINT_WD_IMPL_TAG))) {
         /* we have implicit OR explicit default node */
         /* get with-defaults module */
         wdmod = ly_ctx_get_module_implemented(LYD_CTX(node), "ietf-netconf-with-defaults");
     }
 
-    if (node->schema && node->meta) {
+    if (node->schema && (node->meta || wdmod)) {
         if (inner) {
-            LY_CHECK_RET(json_print_member2(pctx, NULL, LY_VALUE_JSON, NULL, 1));
+            LY_CHECK_RET(json_print_member2(pctx, lyd_parent(node), LY_VALUE_JSON, NULL, 1));
         } else {
             LY_CHECK_RET(json_print_member(pctx, node, 1));
         }
@@ -482,9 +491,9 @@ json_print_attributes(struct jsonpr_ctx *pctx, const struct lyd_node *node, ly_b
         LEVEL_PRINTED;
     } else if (!node->schema && ((struct lyd_node_opaq *)node)->attr) {
         if (inner) {
-            LY_CHECK_RET(json_print_member2(pctx, NULL, LY_VALUE_JSON, NULL, 1));
+            LY_CHECK_RET(json_print_member2(pctx, lyd_parent(node), LY_VALUE_JSON, NULL, 1));
         } else {
-            LY_CHECK_RET(json_print_member2(pctx, node, ((struct lyd_node_opaq *)node)->format,
+            LY_CHECK_RET(json_print_member2(pctx, lyd_parent(node), ((struct lyd_node_opaq *)node)->format,
                     &((struct lyd_node_opaq *)node)->name, 1));
         }
         ly_print_(pctx->out, "{%s", (DO_FORMAT ? "\n" : ""));
@@ -531,15 +540,19 @@ json_print_any_content(struct jsonpr_ctx *pctx, struct lyd_node_any *any)
     LY_ERR ret = LY_SUCCESS;
     struct lyd_node *iter;
     const struct lyd_node *prev_parent;
-    uint32_t prev_opts, prev_lo;
+    uint32_t prev_opts, temp_lo = 0;
 
     assert(any->schema->nodetype & LYD_NODE_ANY);
 
+    if ((any->schema->nodetype == LYS_ANYDATA) && (any->value_type != LYD_ANYDATA_DATATREE)) {
+        LOGINT_RET(pctx->ctx);
+    }
+
     if (any->value_type == LYD_ANYDATA_LYB) {
         uint32_t parser_options = LYD_PARSE_ONLY | LYD_PARSE_OPAQ | LYD_PARSE_STRICT;
 
         /* turn logging off */
-        prev_lo = ly_log_options(0);
+        ly_temp_log_options(&temp_lo);
 
         /* try to parse it into a data tree */
         if (lyd_parse_data_mem(pctx->ctx, any->value.mem, LYD_LYB, parser_options, 0, &iter) == LY_SUCCESS) {
@@ -549,8 +562,8 @@ json_print_any_content(struct jsonpr_ctx *pctx, struct lyd_node_any *any)
             any->value_type = LYD_ANYDATA_DATATREE;
         }
 
-        /* turn loggin on again */
-        ly_log_options(prev_lo);
+        /* turn logging on again */
+        ly_temp_log_options(NULL);
     }
 
     switch (any->value_type) {
@@ -603,7 +616,7 @@ json_print_any_content(struct jsonpr_ctx *pctx, struct lyd_node_any *any)
             }
         } else {
             /* print as a string */
-            ly_print_(pctx->out, "\"%s\"", any->value.str);
+            json_print_string(pctx->out, any->value.str);
         }
         break;
     case LYD_ANYDATA_LYB:
@@ -752,6 +765,8 @@ json_print_array_is_last_inst(struct jsonpr_ctx *pctx, const struct lyd_node *no
 static LY_ERR
 json_print_leaf_list(struct jsonpr_ctx *pctx, const struct lyd_node *node)
 {
+    const struct lys_module *wdmod = NULL;
+
     if (!is_open_array(pctx, node)) {
         LY_CHECK_RET(json_print_member(pctx, node, 0));
         LY_CHECK_RET(json_print_array_open(pctx, node));
@@ -763,20 +778,22 @@ json_print_leaf_list(struct jsonpr_ctx *pctx, const struct lyd_node *node)
     }
 
     if (node->schema->nodetype == LYS_LIST) {
-        if (!lyd_child(node)) {
-            /* empty, e.g. in case of filter */
-            ly_print_(pctx->out, "%s%snull", (pctx->level_printed >= pctx->level) ? "," : "", DO_FORMAT ? " " : "");
-            LEVEL_PRINTED;
-        } else {
-            /* print list's content */
-            LY_CHECK_RET(json_print_inner(pctx, node));
-        }
+        /* print list's content */
+        LY_CHECK_RET(json_print_inner(pctx, node));
     } else {
         assert(node->schema->nodetype == LYS_LEAFLIST);
+
         LY_CHECK_RET(json_print_value(pctx, LYD_CTX(node), &((const struct lyd_node_term *)node)->value));
 
-        if (node->meta && !pctx->print_sibling_metadata) {
-            pctx->print_sibling_metadata = node;
+        if (!pctx->print_sibling_metadata) {
+            if ((node->flags & LYD_DEFAULT) && (pctx->options & (LYD_PRINT_WD_ALL_TAG | LYD_PRINT_WD_IMPL_TAG))) {
+                /* we have implicit OR explicit default node, get with-defaults module */
+                wdmod = ly_ctx_get_module_implemented(LYD_CTX(node), "ietf-netconf-with-defaults");
+            }
+            if (node->meta || wdmod) {
+                /* we will be printing metadata for these siblings */
+                pctx->print_sibling_metadata = node;
+            }
         }
     }
 
@@ -798,26 +815,31 @@ static LY_ERR
 json_print_metadata_leaflist(struct jsonpr_ctx *pctx)
 {
     const struct lyd_node *prev, *node, *iter;
+    const struct lys_module *wdmod = NULL;
 
     if (!pctx->print_sibling_metadata) {
         return LY_SUCCESS;
     }
 
+    if (pctx->options & (LYD_PRINT_WD_ALL_TAG | LYD_PRINT_WD_IMPL_TAG)) {
+        /* get with-defaults module */
+        wdmod = ly_ctx_get_module_implemented(pctx->ctx, "ietf-netconf-with-defaults");
+    }
+
+    /* node is the first instance of the leaf-list */
     for (node = pctx->print_sibling_metadata, prev = pctx->print_sibling_metadata->prev;
             prev->next && matching_node(node, prev);
             node = prev, prev = node->prev) {}
 
-    /* node is the first instance of the leaf-list */
-
     LY_CHECK_RET(json_print_member(pctx, node, 1));
     ly_print_(pctx->out, "[%s", (DO_FORMAT ? "\n" : ""));
     LEVEL_INC;
     LY_LIST_FOR(node, iter) {
         PRINT_COMMA;
-        if (iter->meta) {
+        if (iter->meta || (iter->flags & LYD_DEFAULT)) {
             ly_print_(pctx->out, "%*s%s", INDENT, DO_FORMAT ? "{\n" : "{");
             LEVEL_INC;
-            LY_CHECK_RET(json_print_metadata(pctx, iter, NULL));
+            LY_CHECK_RET(json_print_metadata(pctx, iter, (iter->flags & LYD_DEFAULT) ? wdmod : NULL));
             LEVEL_DEC;
             ly_print_(pctx->out, "%s%*s}", DO_FORMAT ? "\n" : "", INDENT);
         } else {
