@@ -124,7 +124,7 @@ cmd_data(struct ly_ctx **ctx, const char *cmdline)
     uint8_t data_merge = 0;
     uint32_t options_print = 0;
     uint32_t options_parse = YL_DEFAULT_DATA_PARSE_OPTIONS;
-    uint32_t options_validate = 0;
+    uint32_t options_validate = YL_DEFAULT_DATA_VALIDATE_OPTIONS;
     enum lyd_type data_type = 0;
     uint8_t data_type_set = 0;
     LYD_FORMAT outformat = LYD_UNKNOWN;
@@ -227,10 +227,8 @@ cmd_data(struct ly_ctx **ctx, const char *cmdline)
                 options_validate |= LYD_VALIDATE_NO_STATE;
             } else if (!strcasecmp(optarg, "get")) {
                 options_parse |= LYD_PARSE_ONLY;
-            } else if (!strcasecmp(optarg, "getconfig") || !strcasecmp(optarg, "get-config")) {
+            } else if (!strcasecmp(optarg, "getconfig") || !strcasecmp(optarg, "get-config") || !strcasecmp(optarg, "edit")) {
                 options_parse |= LYD_PARSE_ONLY | LYD_PARSE_NO_STATE;
-            } else if (!strcasecmp(optarg, "edit")) {
-                options_parse |= LYD_PARSE_ONLY;
             } else if (!strcasecmp(optarg, "rpc") || !strcasecmp(optarg, "action")) {
                 data_type = LYD_TYPE_RPC_YANG;
             } else if (!strcasecmp(optarg, "reply") || !strcasecmp(optarg, "rpcreply")) {
