@@ -1,9 +1,10 @@
 /**
  * @file main_ni.c
  * @author Radek Krejci <rkrejci@cesnet.cz>
- * @brief libyang's yanglint tool - noninteractive code
+ * @author Michal Vasko <mvasko@cesnet.cz>
+ * @brief libyang's yanglint tool - non-interactive code
  *
- * Copyright (c) 2020 CESNET, z.s.p.o.
+ * Copyright (c) 2020 - 2022 CESNET, z.s.p.o.
  *
  * This source code is licensed under BSD 3-Clause License (the "License").
  * You may not use this file except in compliance with the License.
@@ -24,6 +25,7 @@
 #include <sys/stat.h>
 
 #include "libyang.h"
+#include "plugins_exts.h"
 
 #include "common.h"
 #include "out.h"
@@ -41,7 +43,7 @@ struct context {
     /* prepared output (--output option or stdout by default) */
     struct ly_out *out;
 
-    struct ly_set searchpaths;
+    char *searchpaths;
 
     /* options flags */
     uint8_t list;        /* -l option to print list of schemas */
@@ -67,8 +69,15 @@ struct context {
     const struct lysc_node *schema_node;
     const char *submodule;
 
+    /* name of file containing explicit context passed to callback
+     * for schema-mount extension.  This also causes a callback to
+     * be registered.
+     */
+    char *schema_context_filename;
+
     /* value of --format in case of schema format */
     LYS_OUTFORMAT schema_out_format;
+    ly_bool feature_param_format;
 
     /*
      * data
@@ -107,10 +116,15 @@ erase_context(struct context *c)
     ly_set_erase(&c->schema_modules, NULL);
 
     /* context */
-    ly_set_erase(&c->searchpaths, NULL);
+    free(c->searchpaths);
+    c->searchpaths = NULL;
 
     ly_out_free(c->out, NULL,  0);
     ly_ctx_destroy(c->ctx);
+
+    if (c->schema_context_filename) {
+        free(c->schema_context_filename);
+    }
 }
 
 static void
@@ -152,7 +166,7 @@ help(int shortout)
 
     printf("  -f FORMAT, --format=FORMAT\n"
             "                Convert input into FORMAT. Supported formats: \n"
-            "                yang, yin, tree and info for schemas,\n"
+            "                yang, yin, tree, info and feature-param for schemas,\n"
             "                xml, json, and lyb for data.\n\n");
 
     printf("  -p PATH, --path=PATH\n"
@@ -190,6 +204,11 @@ help(int shortout)
     printf("  -s SUBMODULE, --submodule=SUBMODULE\n"
             "                Print the specific submodule instead of the main module.\n\n");
 
+    printf("  -x FILE, --ext-data=FILE\n"
+            "                File containing the specific data required by an extension. Required by\n"
+            "                the schema-mount extension, for example, when the operational data are\n"
+            "                expected in the file. File format is guessed.\n\n");
+
     printf("  -n, --not-strict\n"
             "                Do not require strict data parsing (silently skip unknown data),\n"
             "                has no effect for schemas.\n\n");
@@ -307,6 +326,7 @@ get_features_not_applied(const struct ly_set *fset)
 {
     for (uint32_t u = 0; u < fset->count; ++u) {
         struct schema_features *sf = fset->objs[u];
+
         if (!sf->applied) {
             return sf;
         }
@@ -315,43 +335,76 @@ get_features_not_applied(const struct ly_set *fset)
     return NULL;
 }
 
+static LY_ERR
+ext_data_clb(const struct lysc_ext_instance *ext, void *user_data, void **ext_data, ly_bool *ext_data_free)
+{
+    struct ly_ctx *ctx;
+    struct lyd_node *data = NULL;
+
+    ctx = ext->module->ctx;
+    if (user_data) {
+        lyd_parse_data_path(ctx, user_data, LYD_XML, LYD_PARSE_STRICT, LYD_VALIDATE_PRESENT, &data);
+    }
+
+    *ext_data = data;
+    *ext_data_free = 1;
+    return LY_SUCCESS;
+}
+
+static LY_ERR
+searchpath_strcat(char **searchpaths, const char *path)
+{
+    uint64_t len;
+    char *new;
+
+    if (!(*searchpaths)) {
+        *searchpaths = strdup(path);
+        return LY_SUCCESS;
+    }
+
+    len = strlen(*searchpaths) + strlen(path) + strlen(PATH_SEPARATOR);
+    new = realloc(*searchpaths, sizeof(char) * len + 1);
+    if (!new) {
+        return LY_EMEM;
+    }
+    strcat(new, PATH_SEPARATOR);
+    strcat(new, path);
+    *searchpaths = new;
+
+    return LY_SUCCESS;
+}
+
 static int
 fill_context_inputs(int argc, char *argv[], struct context *c)
 {
     struct ly_in *in = NULL;
     struct schema_features *sf;
     struct lys_module *mod;
-    const char *all_features[] = {"*", NULL}, *searchdir;
+    const char *all_features[] = {"*", NULL};
     char *dir = NULL, *module = NULL;
 
-    /* get the first searchdir */
-    searchdir = c->searchpaths.count ? c->searchpaths.objs[0] : NULL;
-
+    /* Create libyang context. */
     if (c->yang_lib_file) {
         /* ignore features */
         ly_set_erase(&c->schema_features, free_features);
 
-        /* create context from the yang-library file */
-        if (ly_ctx_new_ylpath(searchdir, c->yang_lib_file, LYD_UNKNOWN, c->ctx_options, &c->ctx)) {
-            YLMSG_E("Unable to create libyang context from yang-library data.\n");
+        if (ly_ctx_new_ylpath(c->searchpaths, c->yang_lib_file, LYD_UNKNOWN, c->ctx_options, &c->ctx)) {
+            YLMSG_E("Unable to modify libyang context with yang-library data.\n");
             return -1;
         }
     } else {
         /* set imp feature flag if all should be enabled */
-        if (!c->schema_features.count) {
-            c->ctx_options |= LY_CTX_ENABLE_IMP_FEATURES;
-        }
+        c->ctx_options |= !c->schema_features.count ? LY_CTX_ENABLE_IMP_FEATURES : 0;
 
-        /* libyang context */
-        if (ly_ctx_new(searchdir, c->ctx_options, &c->ctx)) {
-            YLMSG_E("Unable to create libyang context.\n");
+        if (ly_ctx_new(c->searchpaths, c->ctx_options, &c->ctx)) {
+            YLMSG_E("Unable to create libyang context\n");
             return -1;
         }
+    }
 
-        /* set the rest of searchdirs */
-        for (uint32_t i = 1; i < c->searchpaths.count; ++i) {
-            ly_ctx_set_searchdir(c->ctx, c->searchpaths.objs[i]);
-        }
+    /* set callback providing run-time extension instance data */
+    if (c->schema_context_filename) {
+        ly_ctx_set_ext_data_clb(c->ctx, ext_data_clb, c->schema_context_filename);
     }
 
     /* process the operational and/or reply RPC content if any */
@@ -378,7 +431,6 @@ fill_context_inputs(int argc, char *argv[], struct context *c)
             LY_ERR ret;
             uint8_t path_unset = 1; /* flag to unset the path from the searchpaths list (if not already present) */
             const char **features;
-            struct lys_module *mod;
 
             /* parse the input */
             if (parse_schema_path(argv[optind + i], &dir, &module)) {
@@ -420,7 +472,7 @@ fill_context_inputs(int argc, char *argv[], struct context *c)
             free(module);
             module = NULL;
 
-            if (c->schema_out_format) {
+            if (c->schema_out_format || c->feature_param_format) {
                 /* module will be printed */
                 if (ly_set_add(&c->schema_modules, (void *)mod, 1, NULL)) {
                     YLMSG_E("Storing parsed schema module (%s) for print failed.\n", argv[optind + i]);
@@ -493,6 +545,7 @@ fill_context(int argc, char *argv[], struct context *c)
         {"schema-node",       required_argument, NULL, 'P'},
         {"single-node",       no_argument,       NULL, 'q'},
         {"submodule",         required_argument, NULL, 's'},
+        {"ext-data",          required_argument, NULL, 'x'},
         {"not-strict",        no_argument,       NULL, 'n'},
         {"present",           no_argument,       NULL, 'e'},
         {"type",              required_argument, NULL, 't'},
@@ -514,14 +567,16 @@ fill_context(int argc, char *argv[], struct context *c)
 
     c->ctx_options = YL_DEFAULT_CTX_OPTIONS;
     c->data_parse_options = YL_DEFAULT_DATA_PARSE_OPTIONS;
+    c->data_validate_options = YL_DEFAULT_DATA_VALIDATE_OPTIONS;
     c->line_length = 0;
 
     opterr = 0;
 #ifndef NDEBUG
-    while ((opt = getopt_long(argc, argv, "hvVQf:p:DF:iP:qs:net:d:lL:o:O:R:myY:G:", options, &opt_index)) != -1) {
+    while ((opt = getopt_long(argc, argv, "hvVQf:p:DF:iP:qs:net:d:lL:o:O:R:myY:x:G:", options, &opt_index)) != -1)
 #else
-    while ((opt = getopt_long(argc, argv, "hvVQf:p:DF:iP:qs:net:d:lL:o:O:R:myY:", options, &opt_index)) != -1) {
+    while ((opt = getopt_long(argc, argv, "hvVQf:p:DF:iP:qs:net:d:lL:o:O:R:myY:x:", options, &opt_index)) != -1)
 #endif
+    {
         switch (opt) {
         case 'h': /* --help */
             help(0);
@@ -533,6 +588,7 @@ fill_context(int argc, char *argv[], struct context *c)
 
         case 'V': { /* --verbose */
             LY_LOG_LEVEL verbosity = ly_log_level(LY_LLERR);
+
             if (verbosity < LY_LLDBG) {
                 ++verbosity;
             }
@@ -542,6 +598,7 @@ fill_context(int argc, char *argv[], struct context *c)
 
         case 'Q': { /* --quiet */
             LY_LOG_LEVEL verbosity = ly_log_level(LY_LLERR);
+
             if (verbosity == LY_LLERR) {
                 /* turn logging off */
                 ly_log_options(LY_LOSTORE_LAST);
@@ -574,6 +631,8 @@ fill_context(int argc, char *argv[], struct context *c)
             } else if (!strcasecmp(optarg, "lyb")) {
                 c->schema_out_format = 0;
                 c->data_out_format = LYD_LYB;
+            } else if (!strcasecmp(optarg, "feature-param")) {
+                c->feature_param_format = 1;
             } else {
                 YLMSG_E("Unknown output format %s\n", optarg);
                 help(1);
@@ -593,7 +652,7 @@ fill_context(int argc, char *argv[], struct context *c)
                 return -1;
             }
 
-            if (ly_set_add(&c->searchpaths, optarg, 0, NULL)) {
+            if (searchpath_strcat(&c->searchpaths, optarg)) {
                 YLMSG_E("Storing searchpath failed.\n");
                 return -1;
             }
@@ -640,6 +699,10 @@ fill_context(int argc, char *argv[], struct context *c)
             c->submodule = optarg;
             break;
 
+        case 'x': /* --ext-data */
+            c->schema_context_filename = strdup(optarg);
+            break;
+
         case 'n': /* --not-strict */
             c->data_parse_options &= ~LYD_PARSE_STRICT;
             break;
@@ -875,6 +938,9 @@ main_ni(int argc, char *argv[])
 {
     int ret = EXIT_SUCCESS, r;
     struct context c = {0};
+    char *features_output = NULL;
+    struct ly_set set = {0};
+    uint32_t u;
 
     /* set callback for printing libyang messages */
     ly_set_log_clb(libyang_verbclb, 1);
@@ -893,7 +959,20 @@ main_ni(int argc, char *argv[])
         /* print the list of schemas */
         print_list(c.out, c.ctx, c.data_out_format);
     } else {
-        if (c.schema_out_format) {
+        if (c.feature_param_format) {
+            for (u = 0; u < c.schema_modules.count; u++) {
+                if (collect_features(c.schema_modules.objs[u], &set)) {
+                    YLMSG_E("Unable to read features from a module.\n");
+                    goto cleanup;
+                }
+                if (generate_features_output(c.schema_modules.objs[u], &set, &features_output)) {
+                    YLMSG_E("Unable to generate feature command output.\n");
+                    goto cleanup;
+                }
+                ly_set_erase(&set, NULL);
+            }
+            ly_print(c.out, "%s\n", features_output);
+        } else if (c.schema_out_format) {
             if (c.schema_node) {
                 ret = lys_print_node(c.out, c.schema_node, c.schema_out_format, 0, c.schema_print_options);
                 if (ret) {
@@ -902,6 +981,7 @@ main_ni(int argc, char *argv[])
                 }
             } else if (c.submodule) {
                 const struct lysp_submodule *submod = ly_ctx_get_submodule_latest(c.ctx, c.submodule);
+
                 if (!submod) {
                     YLMSG_E("Unable to find submodule %s.\n", c.submodule);
                     goto cleanup;
@@ -913,7 +993,7 @@ main_ni(int argc, char *argv[])
                     goto cleanup;
                 }
             } else {
-                for (uint32_t u = 0; u < c.schema_modules.count; ++u) {
+                for (u = 0; u < c.schema_modules.count; ++u) {
                     ret = lys_print_module(c.out, (struct lys_module *)c.schema_modules.objs[u], c.schema_out_format,
                             c.line_length, c.schema_print_options);
                     /* for YANG Tree Diagrams printing it's more readable to print a blank line between modules. */
@@ -941,6 +1021,8 @@ main_ni(int argc, char *argv[])
 cleanup:
     /* cleanup */
     erase_context(&c);
+    free(features_output);
+    ly_set_erase(&set, NULL);
 
     return ret;
 }
