@@ -1,11 +1,12 @@
-#! /bin/sh
+#!/usr/bin/env bash
 
 LIST_BASE="List of the loaded models:
     i ietf-yang-metadata@2016-08-05
-    I yang@2021-04-07
+    I yang@2022-06-16
     i ietf-inet-types@2013-07-15
     i ietf-yang-types@2013-07-15
-    I ietf-yang-schema-mount@2019-01-14"
+    I ietf-yang-schema-mount@2019-01-14
+    i ietf-yang-structure-ext@2020-06-17"
 
 testListEmptyContext() {
   output=`${YANGLINT} -l`
@@ -15,10 +16,11 @@ testListEmptyContext() {
 testListAllImplemented() {
   LIST_BASE_ALLIMPLEMENTED="List of the loaded models:
     I ietf-yang-metadata@2016-08-05
-    I yang@2021-04-07
+    I yang@2022-06-16
     I ietf-inet-types@2013-07-15
     I ietf-yang-types@2013-07-15
-    I ietf-yang-schema-mount@2019-01-14"
+    I ietf-yang-schema-mount@2019-01-14
+    I ietf-yang-structure-ext@2020-06-17"
   output=`${YANGLINT} -lii`
   assertEquals "Unexpected list of modules in empty context with -ii." "${LIST_BASE_ALLIMPLEMENTED}" "${output}"
 }
